# Generated by Django 5.2b1 on 2025-03-04 13:37

from django.db import migrations


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("models", "11499_add_editlog_resourceinstance_idx"),  # last 7.6 migration
    ]

    operations = [
        migrations.CreateModel(
            name="GraphWithPrefetching",
            fields=[],
            options={
                "db_table": "graphs",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("models.graphmodel",),
        ),
        migrations.CreateModel(
            name="ResourceTileTree",
            fields=[],
            options={
                "db_table": "resource_instances",
                "permissions": (("no_access_to_resourceinstance", "No Access"),),
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("models.resourceinstance",),
        ),
        migrations.CreateModel(
            name="TileTree",
            fields=[],
            options={
                "db_table": "tiles",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("models.tilemodel",),
        ),
    ]
